
import posixpath
import uuid
import pandas as pd

import tempfile
from pathlib import Path

import pytest
import requests_mock

from giskard import slicing_function, transformation_function
from giskard.ml_worker.core.savable import Artifact
from giskard.ml_worker.testing.test_result import TestResult as GiskardTestResult
from giskard import test

from tests.utils import (
    CALLABLE_FUNCTION_META_CACHE,
    CALLABLE_FUNCTION_PKL_CACHE,
    MockedClient,
    MockedProjectCacheDir,
    fixup_mocked_artifact_meta_version,
    get_local_cache_callable_artifact,
    register_uri_for_artifact_meta_info,
    register_uri_for_artifact_info,
    get_url_for_artifact_meta_info,
    get_url_for_artifacts_base,
    local_save_artifact_under_giskard_home_cache,
    is_url_requested,
)


BASE_CLIENT_URL = "http://giskard-host:12345/api/v2"


# Define a test function
@test
def my_custom_test(model, data):
    return GiskardTestResult(passed=True)


# Define a slicing function
@slicing_function(row_level=False)
def head_slice(df: pd.DataFrame) -> pd.DataFrame:
    return df.head(10)


# Define a transformation function
@transformation_function()
def do_nothing(row):
    return row


def test_download_global_test_function_from_registry():
    cf: Artifact = my_custom_test

    # Load from registry using uuid without client
    download_cf = cf.__class__.download(uuid=cf.meta.uuid, client=None, project_key=None)

    # Check the downloaded info
    assert download_cf.__class__ is cf.__class__
    assert download_cf.meta.uuid == cf.meta.uuid


@pytest.mark.parametrize(
    "cf",
    [
        my_custom_test, # Test
        head_slice,     # Slice
        do_nothing,     # Transformation
    ],
)
def test_download_global_test_function_from_local(cf):
    project_key = str(uuid.uuid4())
    with MockedProjectCacheDir(project_key):
        cf.meta.uuid = str(uuid.uuid4())    # Regenerate a UUID to ensure not loading from registry

        local_save_artifact_under_giskard_home_cache(cf, project_key=None)

        # Load from registry using uuid without client
        download_cf = cf.__class__.download(uuid=cf.meta.uuid, client=None, project_key=None)

        # Check the downloaded info
        assert download_cf.__class__ is cf.__class__
        assert download_cf.meta.uuid == cf.meta.uuid


@pytest.mark.parametrize(
    "cf",
    [
        my_custom_test, # Test
        head_slice,     # Slice
        do_nothing,     # Transformation
    ],
)
def test_download_callable_function(cf: Artifact):
    with MockedClient(mock_all=False) as (client, mr):
        cf.meta.uuid = str(uuid.uuid4())    # Regenerate a UUID to ensure not loading from registry
        cache_dir = get_local_cache_callable_artifact(project_key=None, artifact=cf)

        # Save to temp
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir_path = Path(tmpdir)
            cf.save(tmpdir_path)
            # Check saved file
            assert (tmpdir_path / CALLABLE_FUNCTION_PKL_CACHE).exists()
            assert (tmpdir_path / CALLABLE_FUNCTION_META_CACHE).exists()

            requested_urls = []
            # Fixup the differences from Backend
            meta_info = fixup_mocked_artifact_meta_version(cf.meta.to_json())
            # Fixup the name to avoid load from module
            meta_info.update({
                "name": f"fake_{cf._get_name()}",
            })
            url = get_url_for_artifact_meta_info(cf, project_key=None)
            mr.register_uri(method=requests_mock.GET, url=url, json=meta_info)
            requested_urls.append(url)

            # Register for Artifact info
            requested_urls.extend(register_uri_for_artifact_info(mr, cf, project_key=None))

            # Register for Artifacts content
            artifacts_base_url = get_url_for_artifacts_base(cf, project_key=None)
            for file in [CALLABLE_FUNCTION_META_CACHE, CALLABLE_FUNCTION_PKL_CACHE]:
                with open(tmpdir_path / file, "rb") as f:
                    mr.register_uri(
                        method=requests_mock.GET,
                        url=posixpath.join(artifacts_base_url, file),
                        content=f.read(),
                    )
                    requested_urls.append(posixpath.join(artifacts_base_url, file))

            # Download: should not call load_artifact to request and download
            download_cf = cf.__class__.download(uuid=cf.meta.uuid, client=client, project_key=None)

            for requested_url in requested_urls:
                assert is_url_requested(mr.request_history, requested_url)

            # Check the downloaded info
            assert download_cf.__class__ is cf.__class__
            assert download_cf.meta.uuid == cf.meta.uuid
            # Check the downloaded files
            assert (cache_dir / CALLABLE_FUNCTION_PKL_CACHE).exists()
            assert (cache_dir / CALLABLE_FUNCTION_META_CACHE).exists()


@pytest.mark.parametrize(
    "cf",
    [
        my_custom_test, # Test
        head_slice,     # Slice
        do_nothing,     # Transformation
    ],
)
def test_download_global_callable_function_from_module(cf: Artifact):
    with MockedClient(mock_all=False) as (client, mr):
        cf.meta.uuid = str(uuid.uuid4())    # Regenerate a UUID to ensure not loading from registry
        cache_dir = get_local_cache_callable_artifact(project_key=None, artifact=cf)

        requested_urls = []
        # Prepare global URL
        requested_urls.extend(register_uri_for_artifact_meta_info(mr, cf, project_key=None))

        # Download: should not call load_artifact to request and download
        download_cf = cf.__class__.download(uuid=cf.meta.uuid, client=client, project_key=None)

        for requested_url in requested_urls:
            assert is_url_requested(mr.request_history, requested_url)

        # Check the downloaded info
        assert download_cf.__class__ is cf.__class__
        assert download_cf.meta.uuid == cf.meta.uuid
        # Check the files that do not need to be downloaded
        assert not (cache_dir / CALLABLE_FUNCTION_PKL_CACHE).exists()
        assert not (cache_dir / CALLABLE_FUNCTION_META_CACHE).exists()


@pytest.mark.parametrize(
    "cf",
    [
        my_custom_test, # Test
        head_slice,     # Slice
        do_nothing,     # Transformation
    ],
)
def test_download_global_callable_function_from_cache(cf: Artifact):
    with MockedClient(mock_all=False) as (client, mr):
        cf.meta.uuid = str(uuid.uuid4())    # Regenerate a UUID
        cache_dir = get_local_cache_callable_artifact(project_key=None, artifact=cf)

        # Save to local cache
        local_save_artifact_under_giskard_home_cache(artifact=cf, project_key=None)
        assert (cache_dir / CALLABLE_FUNCTION_PKL_CACHE).exists()
        assert (cache_dir / CALLABLE_FUNCTION_META_CACHE).exists()

        requested_urls = register_uri_for_artifact_meta_info(mr, cf, project_key=None)

        # Download: should not call load_artifact to request and download
        download_cf = cf.__class__.download(uuid=cf.meta.uuid, client=client, project_key=None)

        for requested_url in requested_urls:
            assert is_url_requested(mr.request_history, requested_url)

        # Check the downloaded info
        assert download_cf.__class__ is cf.__class__
        assert download_cf.meta.uuid == cf.meta.uuid



@pytest.mark.parametrize(
    "cf",
    [
        my_custom_test, # Test
        head_slice,     # Slice
        do_nothing,     # Transformation
    ],
)
def test_download_callable_function_in_project(cf: Artifact):
    project_key = str(uuid.uuid4())
    with MockedClient(mock_all=False) as (client, mr), MockedProjectCacheDir(project_key=project_key):
        cf.meta.uuid = str(uuid.uuid4())    # Regenerate a UUID to ensure not loading from registry
        cache_dir = get_local_cache_callable_artifact(project_key=project_key, artifact=cf)

        # Save to temp
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir_path = Path(tmpdir)
            cf.save(tmpdir_path)
            # Check saved file
            assert (tmpdir_path / CALLABLE_FUNCTION_PKL_CACHE).exists()
            assert (tmpdir_path / CALLABLE_FUNCTION_META_CACHE).exists()

            requested_urls = []
            # Fixup the differences from Backend
            meta_info = fixup_mocked_artifact_meta_version(cf.meta.to_json())
            # Fixup the name to avoid load from module
            meta_info.update({
                "name": f"fake_{cf._get_name()}",
            })
            url = get_url_for_artifact_meta_info(cf, project_key=project_key)
            mr.register_uri(method=requests_mock.GET, url=url, json=meta_info)
            requested_urls.append(url)

            # Register for Artifact info
            requested_urls.extend(register_uri_for_artifact_info(mr, cf, project_key=project_key))

            # Register for Artifacts content
            artifacts_base_url = get_url_for_artifacts_base(cf, project_key=project_key)
            for file in [CALLABLE_FUNCTION_META_CACHE, CALLABLE_FUNCTION_PKL_CACHE]:
                with open(tmpdir_path / file, "rb") as f:
                    mr.register_uri(
                        method=requests_mock.GET,
                        url=posixpath.join(artifacts_base_url, file),
                        content=f.read(),
                    )
                    requested_urls.append(posixpath.join(artifacts_base_url, file))

            # Download: should not call load_artifact to request and download
            download_cf = cf.__class__.download(uuid=cf.meta.uuid, client=client, project_key=project_key)

            for requested_url in requested_urls:
                assert is_url_requested(mr.request_history, requested_url)

            # Check the downloaded info
            assert download_cf.__class__ is cf.__class__
            assert download_cf.meta.uuid == cf.meta.uuid
            # Check the downloaded files
            assert (cache_dir / CALLABLE_FUNCTION_PKL_CACHE).exists()
            assert (cache_dir / CALLABLE_FUNCTION_META_CACHE).exists()


@pytest.mark.parametrize(
    "cf",
    [
        my_custom_test, # Test
        head_slice,     # Slice
        do_nothing,     # Transformation
    ],
)
def test_download_callable_function_from_module_in_project(cf: Artifact):
    project_key = str(uuid.uuid4())
    with MockedClient(mock_all=False) as (client, mr), MockedProjectCacheDir(project_key):
        cf.meta.uuid = str(uuid.uuid4())    # Regenerate a UUID to ensure not loading from registry
        cache_dir = get_local_cache_callable_artifact(project_key=project_key, artifact=cf)

        requested_urls = []
        # Prepare global URL
        requested_urls.extend(register_uri_for_artifact_meta_info(mr, cf, project_key))

        # Download: should not call load_artifact to request and download
        download_cf = cf.__class__.download(uuid=cf.meta.uuid, client=client, project_key=project_key)

        for requested_url in requested_urls:
            assert is_url_requested(mr.request_history, requested_url)

        # Check the downloaded info
        assert download_cf.__class__ is cf.__class__
        assert download_cf.meta.uuid == cf.meta.uuid
        # Check the files that do not need to be downloaded
        assert not (cache_dir / CALLABLE_FUNCTION_PKL_CACHE).exists()
        assert not (cache_dir / CALLABLE_FUNCTION_META_CACHE).exists()


@pytest.mark.parametrize(
    "cf",
    [
        my_custom_test, # Test
        head_slice,     # Slice
        do_nothing,     # Transformation
    ],
)
def test_download_callable_function_from_cache_in_project(cf: Artifact):
    project_key = str(uuid.uuid4())
    with MockedClient(mock_all=False) as (client, mr), MockedProjectCacheDir(project_key):
        cf.meta.uuid = str(uuid.uuid4())    # Regenerate a UUID to ensure not loading from registry
        cache_dir = get_local_cache_callable_artifact(project_key=project_key, artifact=cf)

        # Save to local cache
        local_save_artifact_under_giskard_home_cache(artifact=cf, project_key=project_key)
        assert (cache_dir / CALLABLE_FUNCTION_PKL_CACHE).exists()
        assert (cache_dir / CALLABLE_FUNCTION_META_CACHE).exists()

        requested_urls = register_uri_for_artifact_meta_info(mr, cf, project_key)

        # Download: should not call load_artifact to request and download
        download_cf = cf.__class__.download(uuid=cf.meta.uuid, client=client, project_key=project_key)

        for requested_url in requested_urls:
            assert is_url_requested(mr.request_history, requested_url)

        # Check the downloaded info
        assert download_cf.__class__ is cf.__class__
        assert download_cf.meta.uuid == cf.meta.uuid



import pandas as pd
import numpy as np
from pydantic import ValidationError
import pytest
import uuid

import requests_mock

from giskard.datasets.base import Dataset
from giskard.client.dtos import DatasetMetaInfo

from tests import utils
from tests.communications.test_dto_serialization import is_required, get_fields, get_name


# FIXME: conflict on `name` between Giskard Hub (@NotBlank) and Python client (optional in DatasetMeta and DatasetMetaInfo)
MANDATORY_FIELDS = [
    "id",
    "originalSizeBytes",
    "numberOfRows",
    "columnTypes",
    "columnDtypes",
    "compressedSizeBytes",
    "categoryFeatures",
    "createdDate",
]
OPTIONAL_FIELDS = [
    "name",
    "target",
]


valid_df = pd.DataFrame(
    {
        "categorical_column": ["turtle", "crocodile", "turtle"],
        "text_column": ["named Giskard", "a nile crocodile", "etc"],
        "numeric_column": [15.5, 25.9, 2.4],
    }
)
valid_df_column_types = {
    "categorical_column": "category",
    "text_column": "text",
    "numeric_column": "numeric",
}

nonvalid_df = pd.DataFrame(
    {
        "categorical_column": [["turtle"], ["crocodile"], ["turtle"]],
        "text_column": [{1: "named Giskard"}, {2: "a nile crocodile"}, {3: "etc"}],
        "numeric_column": [(15.5, 1), (25.9, 2), (2.4, 3)],
    }
)


def test_factory():
    my_dataset = Dataset(valid_df)
    assert isinstance(my_dataset, Dataset)


def test_valid_df_column_types():
    # Option 0: none of column_types, cat_columns, infer_column_types = True are provided
    with pytest.warns(
        UserWarning,
        match=r"You did not provide the optional argument 'target'\. 'target' is the column name "
        r"in df corresponding to the actual target variable \(ground truth\)\.",
    ):
        my_dataset = Dataset(valid_df)
    assert my_dataset.column_types == {
        "categorical_column": "category",
        "text_column": "text",
        "numeric_column": "numeric",
    }

    # Option 1: column_types is provided
    my_dataset = Dataset(valid_df, column_types=valid_df_column_types)
    assert my_dataset.column_types == valid_df_column_types

    # Option 2: cat_columns is provided
    cat_columns = ["categorical_column"]
    my_dataset = Dataset(valid_df, cat_columns=cat_columns)
    assert my_dataset.column_types == valid_df_column_types

    # Option 3: infer_column_types is provided
    my_dataset = Dataset(valid_df)
    assert my_dataset.column_types == valid_df_column_types


def test_nonvalid_df_column_types():
    # Option 0: none of column_types, cat_columns, infer_column_types = True are provided
    with pytest.raises(
        TypeError,
        match=r"The following columns in your df: \['categorical_column', 'text_column'\] are not hashable\. "
        r"We currently support only hashable column types such as int, bool, str, tuple and not list or dict\.",
    ):
        Dataset(nonvalid_df)

    # Option 1: column_types is provided
    with pytest.raises(
        TypeError,
        match=r"The following columns in your df: \['categorical_column', 'text_column'\] are not hashable\. "
        r"We currently support only hashable column types such as int, bool, str, tuple and not list or dict\.",
    ):
        Dataset(nonvalid_df, column_types=valid_df_column_types)

    # Option 2: cat_columns is provided
    cat_columns = ["categorical_column"]
    with pytest.raises(
        TypeError,
        match=r"The following columns in your df: \['categorical_column', 'text_column'\] are not hashable\. "
        r"We currently support only hashable column types such as int, bool, str, tuple and not list or dict\.",
    ):
        Dataset(nonvalid_df, cat_columns=cat_columns)

    # Option 3: infer_column_types is provided
    with pytest.raises(
        TypeError,
        match=r"The following columns in your df: \['categorical_column', 'text_column'\] are not hashable\. "
        r"We currently support only hashable column types such as int, bool, str, tuple and not list or dict\.",
    ):
        Dataset(nonvalid_df)


def test_dataset_raises_exception_if_mixed_column_types():
    df = pd.DataFrame({"feature": [1, 2, "string", None, np.nan], "target": [0, 0, 1, 1, 0]})

    with pytest.raises(TypeError):
        Dataset(df, target="target")


def test_inference_priority():
    column_types = {"categorical_column": "text"}
    expected_df_column_types = {
        "categorical_column": "text",
        "text_column": "text",
        "numeric_column": "numeric",
    }

    # Case 1: only one column in column_types is provided
    my_dataset = Dataset(valid_df, column_types=column_types)
    assert my_dataset.column_types == expected_df_column_types

    # Case 2: one column in column_types is provided, one in cat_columns
    my_dataset = Dataset(valid_df, column_types=column_types, cat_columns=["categorical_column"])
    assert my_dataset.column_types == valid_df_column_types

    # Case 3: an unknown column in column_types is provided
    column_types = {"unknown_column": "text"}
    my_dataset = Dataset(valid_df, column_types=column_types, cat_columns=["categorical_column"])
    assert my_dataset.column_types == valid_df_column_types


def test_numeric_column_names():
    df = pd.DataFrame(np.ones((10, 3)), columns=[1, 2, 3])

    assert Dataset(df, target=2)
    assert Dataset(df, column_types={1: "numeric"})


def test_infer_column_types():

    # if df_size >= 100     ==> category_threshold = floor(log10(df_size))
    assert Dataset(pd.DataFrame({"f": [1, 2] * 50})).column_types["f"] == "category"
    assert Dataset(pd.DataFrame({"f": ["a", "b"] * 50})).column_types["f"] == "category"
    assert Dataset(pd.DataFrame({"f": ["a", "b"] * 49})).column_types["f"] == "category"
    assert Dataset(pd.DataFrame({"f": [1, 2, 3] * 50})).column_types["f"] == "numeric"
    assert Dataset(pd.DataFrame({"f": ["a", "b", "c"] * 50})).column_types["f"] == "text"

    # if 2 < df_size < 100  ==> category_threshold = 2
    assert Dataset(pd.DataFrame({"f": [1, 2, 1]})).column_types["f"] == "category"
    assert Dataset(pd.DataFrame({"f": ["a", "b", "a"]})).column_types["f"] == "category"
    assert Dataset(pd.DataFrame({"f": [1, 2, 3]})).column_types["f"] == "numeric"
    assert Dataset(pd.DataFrame({"f": ["a", "b", "c"]})).column_types["f"] == "text"

    # if df_size <= 2       ==> category_threshold = 0 (column is text)
    assert Dataset(pd.DataFrame({"f": [1, 2]})).column_types["f"] == "numeric"
    assert Dataset(pd.DataFrame({"f": ["a", "b"]})).column_types["f"] == "text"
    assert Dataset(pd.DataFrame({"f": [1]})).column_types["f"] == "numeric"
    assert Dataset(pd.DataFrame({"f": ["a"]})).column_types["f"] == "text"


def test_dataset_download_with_cache(request):
    dataset: Dataset = request.getfixturevalue("enron_data")
    project_key = str(uuid.uuid4())

    with utils.MockedProjectCacheDir(project_key):
        # Save the dataset to cache dir
        utils.local_save_dataset_under_giskard_home_cache(dataset, project_key=project_key)

        with utils.MockedClient(mock_all=False) as (client, mr):
            # The dataset can be then loaded from the cache, without further requests
            requested_urls = []
            requested_urls.extend(utils.register_uri_for_dataset_meta_info(mr, dataset, project_key))

            downloaded_dataset = Dataset.download(client=client, project_key=project_key, dataset_id=str(dataset.id))

            for requested_url in requested_urls:
                assert utils.is_url_requested(mr.request_history, requested_url)

            assert downloaded_dataset.id == dataset.id
            assert downloaded_dataset.meta == dataset.meta


def test_dataset_download(request):
    dataset: Dataset = request.getfixturevalue("enron_data")
    project_key = str(uuid.uuid4())

    with utils.MockedProjectCacheDir(project_key):
        with utils.MockedClient(mock_all=False) as (client, mr):
            # The dataset needs to request files
            requested_urls = []
            requested_urls.extend(utils.register_uri_for_dataset_meta_info(mr, dataset, project_key))
            requested_urls.extend(
                utils.register_uri_for_dataset_artifact_info(mr, dataset, project_key, register_file_contents=True)
            )

            downloaded_dataset = Dataset.download(client=client, project_key=project_key, dataset_id=str(dataset.id))

            for requested_url in requested_urls:
                assert utils.is_url_requested(mr.request_history, requested_url)

            assert downloaded_dataset.id == dataset.id
            assert downloaded_dataset.meta == dataset.meta


def test_dataset_meta_info():
    klass = DatasetMetaInfo
    mandatory_field_names = []
    optional_field_names = []
    for name, field in get_fields(klass).items():
        mandatory_field_names.append(get_name(name, field)) if is_required(field) else \
            optional_field_names.append(get_name(name, field))
    assert set(mandatory_field_names) == set(MANDATORY_FIELDS)
    assert set(optional_field_names) == set(OPTIONAL_FIELDS)


def test_fetch_dataset_meta(request):
    dataset: Dataset = request.getfixturevalue("enron_data")
    project_key = str(uuid.uuid4())

    for op in OPTIONAL_FIELDS:
        with utils.MockedClient(mock_all=False) as (client, mr):
            meta_info = utils.mock_dataset_meta_info(dataset, project_key)
            meta_info.pop(op)
            mr.register_uri(method=requests_mock.GET, url=utils.get_url_for_dataset(dataset, project_key), json=meta_info)

            # Should not raise
            client.load_dataset_meta(project_key, uuid=str(dataset.id))

    for op in MANDATORY_FIELDS:
        with utils.MockedClient(mock_all=False) as (client, mr):
            meta_info = utils.mock_dataset_meta_info(dataset, project_key)
            meta_info.pop(op)
            mr.register_uri(method=requests_mock.GET, url=utils.get_url_for_dataset(dataset, project_key), json=meta_info)

            # Should raise due to missing of values
            with pytest.raises(ValidationError):
                client.load_dataset_meta(project_key, uuid=str(dataset.id))
